%
% BACKPROPAGATION ALGORITHM
%
dummy = 0;


% sigmoid function
function g = sigmoid(z)
g = 1.0 ./ (1.0 + exp(-z));
end


% derivation of the sigmoid function
function g = sigmoidGradient(z)
a = sigmoid(z);
g = a .* (1-a);
end


% Evaluates a NN (specified by T1 and T2) on a given set of input
% examples X
function Y = FeedForward(X, T1, T2)
  X = [ones(size(X,1),1) X];
  A = sigmoid(X * T1);
  A = [ones(size(A,1),1) A];
  Y = sigmoid(A * T2);
end

% creates an NN layer (Theta matrix) with a given number of inputs
% (fan_in) and a given number of outputs (fan_out). the fan_in value
% should not include the bias node (+1); it is added automatically.
function T = randomNNLayer(fan_in, fan_out, epsilon_init)
  if ! exist("epsilon_init")
    epsilon_init = 0.010;
  end
  T = rand(fan_in+1,fan_out) * 2 * epsilon_init - epsilon_init;
end

% calculates the error of a NN for a given set of input (X) and output
% (Y) examples. the NN is specified with the Theta matrices T1 and T2.
function err = Err(T1, T2, X, Y)
  m = size(X,1);
  Y_predict = FeedForward(X, T1, T2);

   err = 1/m * sum(sum(-Y .* log(Y_predict) - (1-Y) .* log(1 - Y_predict)));
end

function [n p] = classificationError(X, Y, T1, T2)
  m = size(X,1);
  Y_predict = FeedForward(X, T1, T2);
  [W IW] = max(Y');
  [W_predict IW_predict] = max(Y_predict');
  n = m - sum(IW == IW_predict);
  p = n/m;
end


% backpropagation algorithm
function [D1 D2] = Backpropagation(T1, T2, X, Y)
 m = size(X,1);
 A1 = [ones(size(X,1),1) X];
 C2 = A1 * T1;
 A2 = [ones(size(C2,1),1) sigmoid(C2)];
 C3 = A2 * T2;
 A3 = sigmoid(C3);

 delta3 = A3 - Y;

 delta2 = (T2 * delta3')';
 delta2 = delta2(:,2:end);
 delta2 = delta2 .* sigmoidGradient(C2);

 D2 = 1/m * A2' * delta3;
 D1 = 1/m * A1' * delta2;
end

function [T1 T2] = gradientDescend(X, Y, T1, T2, iterations, alpha)

  if ! exist("alpha")
    alpha = 0.100;
  end

  if ! exist("iterations")
    iterations = 100;
  end

  for i = 1:iterations
    [D1 D2] = Backpropagation(T1, T2, X, Y);
    T1 = T1 - alpha*D1;
    T2 = T2 - alpha*D2;
  end

end


% Input data
load train-data.mat;
load test-data.mat;

sizes = 1:20;

for i = 1:size(sizes,2)
m = size(X_train,1);
n = size(X_train,2);
k = size(Y_train,2);
hidden_layer_size = sizes(i);

T1 = randomNNLayer(n,hidden_layer_size);
T2 = randomNNLayer(hidden_layer_size,k);

[T1 T2] = gradientDescend(X_train, Y_train, T1, T2);

[n_train p_train] = classificationError(X_train, Y_train, T1, T2);
[n_test p_test] = classificationError(X_test, Y_test, T1, T2);

printf("%2i %4i (%3.2f%%) %4i (%3.2f%%)\n", hidden_layer_size, n_train, p_train*100, n_test, p_test*100);

end
